import time
import threading

class RateLimiter:
    def __init__(self, max_requests, time_window):
        """
        Initialize the rate limiter.

        :param max_requests: The maximum number of requests allowed within the time window.
        :param time_window: The time window in seconds.
        """
        self.max_requests = max_requests
        self.time_window = time_window
        self.request_count = 0
        self.lock = threading.Lock()
        self.reset_time = time.time()

    def allow_request(self):
        """
        Check if the rate limit has been exceeded. Returns True if the request is allowed, False otherwise.
        """
        with self.lock:
            current_time = time.time()
            if current_time - self.reset_time > self.time_window:
                self.request_count = 0
                self.reset_time = current_time
            if self.request_count < self.max_requests:
                self.request_count += 1
                return True
            else:
                return False

    def reset(self):
        """
        Reset the rate limiter.
        """
        with self.lock:
            self.request_count = 0
            self.reset_time = time.time()



class Track:
    def __init__(self, name, mbid, url, image, streamable, artist, album, **kwargs):
        self.name = name
        self.mbid = mbid
        self.url = url
        self.image = image
        self.streamable = streamable
        self.artist = artist
        self.album = album
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"Track(name={self.name}, artist={self.artist.name}, album={self.album.title})"



class LastFMError(Exception):
    pass

class RateLimitExceeded(LastFMError):
    pass

class InvalidAPIKey(LastFMError):
    pass

class ArtistNotFoundError(LastFMError):
    pass

class AlbumNotFoundError(LastFMError):
    pass

class TrackNotFoundError(LastFMError):
    pass

class UserNotFoundError(LastFMError):
    pass



from .handler import Handler
from .exceptions import *

__all__ = ['Handler', 'LastFMError', 'RateLimitExceeded', 'InvalidAPIKey', 'ArtistNotFoundError', 'AlbumNotFoundError', 'TrackNotFoundError', 'UserNotFoundError']

__version__ = '1.0.0'
__author__ = 'Claqz'
__license__ = 'MIT'



class Event:
    def __init__(self, id, title, artists, venue, location, date, image, url, attendees, description, **kwargs):
        self.id = id
        self.title = title
        self.artists = artists
        self.venue = venue
        self.location = location
        self.date = date
        self.image = image
        self.url = url
        self.attendees = attendees
        self.description = description
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"Event(title={self.title}, artists={self.artists}, venue={self.venue})"



class User:
    def __init__(self, name, url, image, country, age, gender, subscriber, playcount, playlists, bootstrap, registered, type, **kwargs):
        self.name = name
        self.url = url
        self.image = image
        self.country = country
        self.age = age
        self.gender = gender
        self.subscriber = subscriber
        self.playcount = playcount
        self.playlists = playlists
        self.bootstrap = bootstrap
        self.registered = registered
        self.type = type
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"User(name={self.name}, url={self.url})"



import requests
import xmltodict
import logging
import json
from .exceptions import RateLimitExceeded, InvalidAPIKey, ArtistNotFoundError, AlbumNotFoundError, TrackNotFoundError, LastFMError
from .ratelimiter import RateLimiter
from .track import Track
from .album import Album
from .user import User
from .event import Event
from .artist import Artist
from .shout import Shout


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class Handler:
    def __init__(self, api_key):
        self.api_key = api_key
        self.rate_limiter = RateLimiter(max_requests=20, time_window=20)

    def _make_request(self, method, url, params=None):
        if not self.rate_limiter.allow_request():
            raise RateLimitExceeded("Rate limit exceeded")

        try:
            response = requests.request(method, url, params=params, timeout=10)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error(f"HTTP error occurred: {http_err}")
            raise
        except requests.exceptions.RequestException as req_err:
            logger.error(f"Request error occurred: {req_err}")
            raise

        logger.info(f"Response Status Code: {response.status_code}")
        logger.debug(f"Response Text: {response.text}")

        try:
            response_dict = xmltodict.parse(response.text)

            if not isinstance(response_dict, dict):
                raise ValueError("Parsed response is not a dictionary")

            response_json = json.dumps(response_dict)
        except Exception as e:
            logger.error(f"Error parsing XML or converting to JSON: {e}")
            raise LastFMError(f"Error processing response: {e}")

        return response_json

    def _parse_response(self, response_json, *keys):
        try:
            response_dict = json.loads(response_json)
            
            
            if 'lfm' in response_dict and 'error' in response_dict['lfm']:
                error_code = response_dict['lfm']['error']['#text']
                if error_code == '2':
                    raise InvalidAPIKey("Invalid API Key")
                elif error_code == '6':
                    raise ArtistNotFoundError("Artist not found")
                elif error_code == '9':
                    raise AlbumNotFoundError("Album not found")
                elif error_code == '10':
                    raise TrackNotFoundError("Track not found")
                else:
                    raise LastFMError(f"LastFM Error: {error_code}")

            
            data = response_dict
            for key in keys:
                data = data.get(key, {})
            if not data:
                raise LastFMError("Information not found in the response")
            return data
        except json.JSONDecodeError as e:
            logger.error(f"Error decoding JSON: {e}")
            raise LastFMError(f"Error processing response: {e}")

    def get_track(self, track_name, artist_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'track.getinfo',
            'track': track_name,
            'artist': artist_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        track_info = self._parse_response(response_json, 'lfm', 'track')
        return Track(**track_info)

    def get_user(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getinfo',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        user_info = self._parse_response(response_json, 'lfm', 'user')
        return User(**user_info)

    def get_top_artists(self, user_name, period="overall"):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.gettopartists',
            'user': user_name,
            'api_key': self.api_key,
            'period': period,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        top_artists_info = self._parse_response(response_json, 'lfm', 'topartists', 'artist')
        return [Artist(**artist) for artist in top_artists_info]

    def get_top_albums(self, user_name, period="overall"):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.gettopalbums',
            'user': user_name,
            'api_key': self.api_key,
            'period': period,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        top_albums_info = self._parse_response(response_json, 'lfm', 'topalbums', 'album')
        return [Album(**album) for album in top_albums_info]

    def get_top_tracks(self, user_name, period="overall"):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.gettoptracks',
            'user': user_name,
            'api_key': self.api_key,
            'period': period,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        top_tracks_info = self._parse_response(response_json, 'lfm', 'toptracks', 'track')
        return [Track(**track) for track in top_tracks_info]

    def get_friends(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getfriends',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        friends_info = self._parse_response(response_json, 'lfm', 'friends', 'user')
        return [User(**friend) for friend in friends_info]

    def get_neighbours(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getneighbours',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        neighbours_info = self._parse_response(response_json, 'lfm', 'neighbours', 'user')
        return [User(**neighbour) for neighbour in neighbours_info]

    def get_now_playing(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getnowplaying',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        now_playing_info = self._parse_response(response_json, 'lfm', 'nowplaying', 'track')
        return Track(**now_playing_info) if now_playing_info else None

    def get_recent_tracks(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getrecenttracks',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        recent_tracks_info = self._parse_response(response_json, 'lfm', 'recenttracks', 'track')
        return [Track(**track) for track in recent_tracks_info]

    def get_loved_tracks(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getlovedtracks',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        loved_tracks_info = self._parse_response(response_json, 'lfm', 'lovedtracks', 'track')
        return [Track(**track) for track in loved_tracks_info]

    def get_attended_events(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getattendedevents',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        attended_events_info = self._parse_response(response_json, 'lfm', 'attendedevents', 'event')
        return [Event(**event) for event in attended_events_info]

    def get_shouts(self, user_name):
        url = "https://ws.audioscrobbler.com/2.0/"
        params = {
            'method': 'user.getshouts',
            'user': user_name,
            'api_key': self.api_key,
            'format': 'xml'
        }
        response_json = self._make_request('GET', url, params)
        shouts_info = self._parse_response(response_json, 'lfm', 'shouts', 'shout')
        return [Shout(**shout) for shout in shouts_info]



class Album:
    def __init__(self, title, mbid, url, image, artist, **kwargs):
        self.title = title
        self.mbid = mbid
        self.url = url
        self.image = image
        self.artist = artist
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"Album(title={self.title}, artist={self.artist.name})"



from setuptools import setup, find_packages

setup(
    name='pyscrobble',
    version='0.0.4',
    description='A Python package for interacting with the Last.fm API',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Claqz',
    author_email='lastfm@gmail.com',
    packages=find_packages(),
    install_requires=[
        'requests>=2.25.1',
        'xmltodict>=0.12.0',
    ],
    extras_require={
        'dev': [
            'pytest>=6.2.5',
            'flake8>=3.9.2',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    python_requires='>=3.9',
)



class Shout:
    def __init__(self, body, author, date, **kwargs):
        self.body = body
        self.author = author
        self.date = date
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"Shout(body={self.body}, author={self.author})"
